import socket
import struct
import urllib.parse

from pyper.common import pyperDebug


def getUrlRedirect(originalUrl: str, originalUrlPort: int = 60, url: str = "/") -> str:
    """Gets the url to redirect to.

    :param originalUrl: The url you are currently on
    :type originalUrl: str
    :param originalUrlPort: The port you are currently on
    :type originalUrlPort: int, defaults to 60
    :param url: The url originalUrl is telling you to redirect to
    :type url: str
    """
    if url.startswith("piper://"):
        return connectFromUrl(url)
    else:
        return connectFromUrl(
            "piper://" + originalUrl + ":" + str(originalUrlPort) + ("/" + url)
        )


def getHostname(ip: str, port: int, path: str, args: dict = {}) -> str:
    """Formats a hostname from the ip, port, path, and args

    :param ip: The hostname/IP of the server
    :type ip: str
    :param port: The server port, defaults to 60
    :type port: int
    :param path: The path to go to on the server, defaults to /
    :type path: str
    :param args: The args to send to the server, defaults to {}
    :type args: dict

    :rtype: str
    :return: hostname"""
    hostname = "piper://" if not ip.startswith("piper://") else ""
    hostname += ip
    hostname += ":" + str(port)
    if path not in ["/", ""]:
        if not path.startswith("/"):
            hostname += "/"
        hostname += path
    if args:
        hostname += "?" + urllib.parse.urlencode(args)
    return hostname


def connectFromUrl(url: str) -> str:
    """Gets the args for pyper.client.connect (execpt for redirectsAllowed) from a url.

    :param url: The URL
    :type url: str

    :rtype: (str, int, str, dict)
    :return: (ip, port, path, options)"""
    url = url if url.startswith("piper://") else "piper://" + url
    data = urllib.parse.urlparse(url)
    server = data.netloc
    port, path, options = 60, "/", dict()
    if server.find("[", 0, 2) == -1:
        if server.find(":", 0, len(server)) == -1:
            ip = server
        else:
            ip, port = server.split(":")
    else:
        # IPv6 support
        index = server.find("]", 0, len(server))
        if index == -1:
            new_ip = "localhost"
        else:
            if server.find(":", index, len(server)) == -1:
                new_ip = server
            else:
                new_ip, port = server.rsplit(":", 1)
        ip = new_ip.strip("[]")

    try:
        port = int(port)
    except ValueError:
        port = 60
    path = data.path if data.path else ""
    options = dict(urllib.parse.parse_qsl(urllib.parse.urlsplit(url).query))
    args = {"ip": ip, "port": port, "path": path, "args": options}
    return args


def connect(
    ip: str,
    port: int = 60,
    path: str = "/",
    args: dict = dict(),
    redirectsAllowed: int = 10,
):
    """Connects to a piper server and return the ID and data of the server reponse

    :param ip: The hostname/IP of the server
    :type ip: str
    :param port: The server port, defaults to 60
    :type port: int
    :param path: The path to go to on the server, defaults to /
    :type path: str
    :param args: The args to send to the server, defaults to {}
    :type args: dict
    :param redirectsAllowed: A number of the amount of redirects allowed, if it is -1 then infinite redirects are allowed, defaults to 10
    :type redirectsAllowed: int

    :rtype: (str, Union[str, bytes])
    :return: (ID, content)
    """
    hostname = getHostname(ip, port, path, args)
    # Remove the "piper://hostname:port" part
    hostname = hostname[hostname.index(str(port)) + len(str(port)) :]
    # Connect the socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((ip, port))
    # Pack the request
    request = struct.pack(
        "<H%ds" % len(hostname), len(hostname), hostname.encode("utf-8")
    )
    if pyperDebug:
        print(request)
    sock.sendall(request)
    # Receive the data
    received = b""
    while (data := sock.recv(4096)) != b"":
        received += data
    try:
        # Unpack the data
        receivedUnpacked = struct.unpack(
            "<Q%ds" % struct.unpack("<Q", received[1:9])[0], received[1:]
        )[1]
        if pyperDebug:
            print(received)
    except:
        return "F3", received
    # Set the default message
    msg = (
        "The data was fomatted incorrectly or an error occurred, here is the raw data: "
        + str(received)
    )
    id = received[0:1]
    # Don't let the server send client side content types
    if id.hex().upper()[1] == "F":
        return "F3", received
    if id in b"\x00\x01":
        # Gemtext and unicode
        msg = receivedUnpacked.decode("utf-8")
    elif id == b"\x02":
        # ASCII
        msg = receivedUnpacked.decode("ascii")
    elif id == b"\x10":
        # File transfer
        msg = receivedUnpacked
    elif id == b"\x20":
        # Redirect
        newUrl = receivedUnpacked.decode("utf-8")
        msg = getUrlRedirect(ip, port, newUrl)
        if redirectsAllowed:
            id, msg = connect(**msg, redirectsAllowed=redirectsAllowed - 1)
        else:
            msg = getHostname(**msg)
    elif id == b"\x22":
        # Resource not found
        msg = "0x22: Resource not found. :("
    elif id == b"\x23":
        # Internal server error
        msg = "0x23: Internal server error."
    elif id == b"\x24":
        # Wrong spec version
        msg = "0x24: Wrong Specification Version."
    # Use .hex() to return the hex as a string. ("aa" instead of b"\xaa")
    return id.hex().upper(), msg



import socketserver
import struct
import urllib.parse
from urllib.parse import urlparse

from pyper.common import pyperDebug

pathToFunc = {}


class PyperServer:
    def __init__(self, port=60):
        """Takes the ip and port of the server and binds the pyper server to the ip and port

        :param ip: The ip to bind the server to
        :type path: str
        :param port: The port to bind the server to, defualts to 60
        :type port: int"""
        self.port = port
        self.serverClass = PyperTCPServer
        self.server = socketserver.TCPServer(("0.0.0.0", port), self.serverClass)
        self.server.allow_reuse_address = True

    def start(self):
        """Starts the pyper server"""
        print("Started server on 0.0.0.0:" + str(self.port) + ".")
        print("Ctrl+C to quit.")
        try:
            with self.server:
                self.server.serve_forever()
        except KeyboardInterrupt:
            # Backspace over the "^C" text
            print("\b\b", end="")
            # Close the socket.
            # TODO: Fix the bug where the socket doesn't close.
            self.server.shutdown()
            self.server.server_close()

    def parseRequest(self, path):
        """The wrapper factory that allows function to be mapped to piper url paths, not meant to be called directly, meant to be used as a decorator

        :param path: The piper url path to be mapped to the function
        :type path: str

        :rtype: function
        :return: wrapper"""

        def wrapper(func):
            """The wrapper itself"""
            pathToFunc.update({path: func})
            # Leaves the function as it is.
            return func

        return wrapper


class PyperTCPServer(socketserver.BaseRequestHandler):
    def handle(self):
        """Handles a connection"""
        self.data = self.request.recv(4096)
        self.fmtAddress = ":".join(map(str, self.client_address))
        self.error = None

        if pyperDebug:
            print(self.data)
        # Unpacks the URL in the request
        try:
            unpackedRequest = struct.unpack(
                "<H%ds" % struct.unpack("<H", self.data[0:2])[0], self.data
            )[1].decode()
        except:
            self.request.close()
            return
        # Set path
        path = urlparse(unpackedRequest).path
        path = path if path else "/"
        # Log the request
        print("<" + self.fmtAddress + ">", path)
        # Set the default to 0x22 (Not found)
        self.contentType, self.returndata = "22", b""
        # Turn the url options into a dict
        options = dict(
            urllib.parse.parse_qsl(urllib.parse.urlsplit(unpackedRequest).query)
        )
        if path in pathToFunc.keys() or "*" in pathToFunc.keys():
            try:
                if path in pathToFunc.keys():
                    self.contentType, self.returndata = pathToFunc[path](
                        {
                            "client_addr": self.client_address,
                            "options": options,
                            "path": path,
                        }
                    )
                # Allows falling back to "*" if a path isn't mapped to a function
                elif "*" in pathToFunc.keys():
                    self.contentType, self.returndata = pathToFunc["*"](
                        {
                            "client_addr": self.client_address,
                            "options": options,
                            "path": path,
                        }
                    )
            except Exception as e:
                self.error = e
                self.contentType, self.returndata = "23", b""

        self.contentType = bytes.fromhex(self.contentType)
        # Encode the data to bytes
        self.returnDataEncoded = self.returndata
        if self.contentType in b"\x00\x01\x20":
            self.returnDataEncoded = bytes(self.returndata, "utf-8")
        elif self.contentType == b"\x02":
            self.returnDataEncoded = bytes(self.returndata, "ascii")
        elif self.contentType == b"\x22\x23":
            # 0x22 and 0x23 don't need data, until they do pyper will force them to use no data
            self.returnDataEncoded = b""

        # Pack the data
        self.rawData = struct.pack(
            "<Q%ds" % len(self.returnDataEncoded),
            len(self.returnDataEncoded),
            self.returnDataEncoded,
        )
        # Add the content type
        self.rawData = self.contentType + self.rawData
        if pyperDebug:
            print(self.rawData)
        self.request.send(self.rawData)
        self.request.close()
        if self.error:
            raise self.error



"""Pyper is a python implementation of the piper networking protocol.
Piper: https://github.com/Luminoso-256/piper
Pyper: https://github.com/Bigjango13/pyper"""
from pyper import client, common, server

__all__ = ("client", "server", "commom")



from os import environ

# Misc
pyperDebug = bool(environ.get("PYPER_DEBUG"))
piperSpecVersion = "0.8.0"

# Content types
utf8 = "00"
gemtext = "01"
ascii = "02"
file = "10"
redirect = "20"
notFound = "22"
internalError = "23"
specVersion = "24"

# Content types borrowed from libpiper
clientOutOfMemory = "F0"
clientConnectionError = "F1"
clientInternalError = "F2"

# Extra one I added myself
clientInvalidData = "F3"

# Content Type dict
contentType = {
    utf8: "UTF8",
    gemtext: "Gemtext",
    ascii: "ASCII",
    file: "File",
    redirect: "Redirect",
    notFound: "Content not found",
    internalError: "Internal server error",
    specVersion: "Diffrent specification version. (Currently on version "
    + piperSpecVersion
    + ")",
    clientOutOfMemory: "Out of memory",
    clientConnectionError: "Connection error",
    clientInternalError: "Internal client error",
    clientInvalidData: "Invalid data",
}



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="pyper-piper",
    version="0.1.6",
    description="The Piper networking protocol in Python.",
    url="https://github.com/Bigjango13/Pyper",
    author="Bigjango13",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["pyper"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Communications",
        "Topic :: Communications :: File Sharing",
        "Topic :: Internet",
        "Topic :: System :: Networking",
    ],
)
